@@ -12,8 +12,9 @@ class InstalledApp:
     necessary if the app needs to be moved.
     """
 
-    def __init__(self, tbfh, address, app_binary=None):
-        self.tbfh = tbfh  # A `tbfh` object representing this app's header.
+    def __init__(self, tbfh, tbff, address, app_binary=None):
+        self.tbfh = tbfh  # A `TBFHeader` object representing this app's header.
+        self.tbff = tbff  # A `TBFFooter` object representing this app's footer.
         self.app_binary = app_binary  # A binary array of the app _after_ the header.
         self.address = address  # Where on the board this app currently is.
 
@@ -31,6 +32,12 @@ class InstalledApp:
         """
         return self.tbfh.get_app_name()
 
+    def get_app_version(self):
+        """
+        Return the version number stored in a program header.
+        """
+        return self.tbfh.get_app_version()
+
     def is_app(self):
         """
         Whether this is an app or padding.
@@ -84,6 +91,14 @@ class InstalledApp:
         """
         return self.tbfh.has_fixed_addresses()
 
+    def filter_fixed_ram_address(self, ram_address):
+        """
+        Specify the start of RAM to filter TBFs in this TAB. For installed apps
+        this is a no-op because presumably we only installed apps that have a
+        reasonable RAM parameter.
+        """
+        pass
+
     def get_fixed_addresses_flash_and_sizes(self):
         """
         Return a list of tuples of all addresses in flash this app is compiled
@@ -192,6 +207,14 @@ class InstalledApp:
         """
         return self.address
 
+    def verify_credentials(self, public_keys):
+        """
+        Using an optional array of public_key binaries, try to check any
+        contained credentials to verify they are valid.
+        """
+        integrity_blob = self.tbfh.get_binary() + self.app_binary
+        self.tbff.verify_credentials(public_keys, integrity_blob)
+
     def has_app_binary(self):
         """
         Whether we have the actual application binary for this app.
@@ -234,7 +257,6 @@ class InstalledApp:
             or self.app_binary_modified
             or address != self.address
         ):
-
             # Get the actual full app binary.
             binary = self.tbfh.get_binary() + self.app_binary
 
@@ -273,13 +295,22 @@ class InstalledApp:
 
         out = ""
         out += "Name:                  {}\n".format(self.get_name())
+        out += "Version:               {}\n".format(self.get_app_version())
         out += "Enabled:               {}\n".format(self.tbfh.is_enabled())
         out += "Sticky:                {}\n".format(self.tbfh.is_sticky())
         out += "Total Size in Flash:   {} bytes\n".format(self.get_size())
 
         if verbose:
             out += "Address in Flash:      {:#x}\n".format(offset)
-            out += textwrap.indent(str(self.tbfh), "  ")
+            out += textwrap.indent(self.tbfh.to_str_at_address(offset), "  ")
+            if self.tbff:
+                out += "\n"
+                out += textwrap.indent(
+                    self.tbff.to_str_at_address(
+                        offset + self.tbfh.get_binary_end_offset()
+                    ),
+                    "  ",
+                )
         return out
 
     def object(self):
@@ -293,6 +324,7 @@ class InstalledApp:
             "size": self.get_size(),
             "address": self.address,
             "header": self.tbfh.object(),
+            "footer": self.tbff.object(),
         }
 
     def __str__(self):
