@@ -13,7 +13,7 @@ class TabTbf:
     This correlates to a specific .tbf file storing a .tab file.
     """
 
-    def __init__(self, filename, tbfh, binary):
+    def __init__(self, filename, tbfh, binary, tbff):
         """
         - `filename` is the identifier used in the .tab.
         - `tbfh` is the header object
@@ -22,6 +22,7 @@ class TabTbf:
         self.filename = filename
         self.tbfh = tbfh
         self.binary = binary
+        self.tbff = tbff
 
 
 class TabApp:
@@ -51,11 +52,16 @@ class TabApp:
 
         self.tbfs = tbfs  # A list of TabTbfs.
 
+        # Address where RAM for apps starts on the board. This is useful for
+        # filtering TBFs that are fixed address to remove TBFs which have
+        # potentially good flash addresses but wrong RAM addresses.
+        self.ram_address_filter = None
+
     def get_name(self):
         """
         Return the app name.
         """
-        app_names = set([tbf.tbfh.get_app_name() for tbf in self.tbfs])
+        app_names = set([tbf.tbfh.get_app_name() for tbf in self._get_tbfs()])
         if len(app_names) > 1:
             raise TockLoaderException("Different names inside the same TAB?")
         elif len(app_names) == 0:
@@ -76,15 +82,23 @@ class TabApp:
         Mark this app as "sticky" in the app's header. This makes it harder to
         accidentally remove this app if it is a core service or debug app.
         """
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             tbf.tbfh.set_flag("sticky", True)
 
     def get_header(self):
         """
         Return a header if there is only one.
         """
-        if len(self.tbfs) == 1:
-            return self.tbfs[0].tbfh
+        if len(self._get_tbfs()) == 1:
+            return self._get_tbfs()[0].tbfh
+        return None
+
+    def get_footers(self):
+        """
+        Return the footers if there are any.
+        """
+        if len(self._get_tbfs()) == 1:
+            return self._get_tbfs()[0].tbff
         return None
 
     def get_size(self):
@@ -93,17 +107,28 @@ class TabApp:
 
         This is only valid if there is only one TBF.
         """
-        if len(self.tbfs) == 1:
-            return self.tbfs[0].tbfh.get_app_size()
+        if len(self._get_tbfs()) == 1:
+            return self._get_tbfs()[0].tbfh.get_app_size()
         else:
             raise TockLoaderException("Size only valid with one TBF")
 
+    def get_app_version(self):
+        """
+        Return the version number stored in a program header.
+
+        This is only valid if there is only one TBF.
+        """
+        if len(self._get_tbfs()) == 1:
+            return self._get_tbfs()[0].tbfh.get_app_version()
+        else:
+            raise TockLoaderException("Version number only valid with one TBF")
+
     def set_size(self, size):
         """
         Force the entire app to be a certain size. If `size` is smaller than the
         actual app an error will be thrown.
         """
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             header_size = tbf.tbfh.get_header_size()
             binary_size = len(tbf.binary)
             current_size = header_size + binary_size
@@ -120,7 +145,7 @@ class TabApp:
         Force each version of the entire app to be a certain size. If `size` is
         smaller than the actual app nothing happens.
         """
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             header_size = tbf.tbfh.get_header_size()
             binary_size = len(tbf.binary)
             current_size = header_size + binary_size
@@ -139,7 +164,7 @@ class TabApp:
         """
         if constraint == "powers_of_two":
             # Make sure the total app size is a power of two.
-            for tbf in self.tbfs:
+            for tbf in self._get_tbfs():
                 current_size = tbf.tbfh.get_app_size()
                 if (current_size & (current_size - 1)) != 0:
                     # This is not a power of two, but should be.
@@ -155,7 +180,7 @@ class TabApp:
         elif type(constraint) is tuple:
             if constraint[0] == "multiple":
                 size_multiple = constraint[1]
-                for tbf in self.tbfs:
+                for tbf in self._get_tbfs():
                     current_size = tbf.tbfh.get_app_size()
                     if (current_size % size_multiple) != 0:
                         # This is not a multiple of the proper size, but should
@@ -175,12 +200,20 @@ class TabApp:
         address.
         """
         has_fixed_addresses = False
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             if tbf.tbfh.has_fixed_addresses():
                 has_fixed_addresses = True
                 break
         return has_fixed_addresses
 
+    def filter_fixed_ram_address(self, ram_address):
+        """
+        Specify the start of RAM to filter TBFs in this TAB. TBFs with fixed RAM
+        addresses that are not reasonably able to fit with the available RAM are
+        ignored from the TAB.
+        """
+        self.ram_address_filter = ram_address
+
     def get_fixed_addresses_flash_and_sizes(self):
         """
         Return a list of tuples of all addresses in flash this app is compiled
@@ -189,7 +222,7 @@ class TabApp:
         [(address, size), (address, size), ...]
         """
         apps_in_flash = []
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             apps_in_flash.append(
                 (tbf.tbfh.get_fixed_addresses()[1], tbf.tbfh.get_app_size())
             )
@@ -205,7 +238,7 @@ class TabApp:
             return True
 
         # Otherwise, see if we have a TBF which can go at the requested address.
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             fixed_flash_address = tbf.tbfh.get_fixed_addresses()[1]
             tbf_header_length = tbf.tbfh.get_header_size()
 
@@ -259,7 +292,7 @@ class TabApp:
         # Find the binary with the lowest valid address that is above `address`.
         best_address = None
         best_index = None
-        for i, tbf in enumerate(self.tbfs):
+        for i, tbf in enumerate(self._get_tbfs()):
             fixed_flash_address = tbf.tbfh.get_fixed_addresses()[1]
 
             # Align to get a reasonable address for this app.
@@ -274,25 +307,66 @@ class TabApp:
                     best_index = i
 
         if best_index != None:
-            self.tbfs = [self.tbfs[best_index]]
+            self.tbfs = [self._get_tbfs()[best_index]]
             return best_address
         else:
             return None
 
-    def delete_tbfh_tlv(self, tlvid):
+    def delete_tlv(self, tlvid):
         """
-        Delete a particular TLV from each TBF header.
+        Delete a particular TLV from each TBF header and footer.
         """
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             tbf.tbfh.delete_tlv(tlvid)
+            tbf.tbff.delete_tlv(tlvid)
 
     def modify_tbfh_tlv(self, tlvid, field, value):
         """
         Modify a particular TLV from each TBF header to set field=value.
         """
-        for tbf in self.tbfs:
+        for tbf in self._get_tbfs():
             tbf.tbfh.modify_tlv(tlvid, field, value)
 
+    def add_tbfh_tlv(self, tlvid, parameters):
+        """
+        Add a particular TLV to each TBF's header.
+        """
+        for tbf in self._get_tbfs():
+            tbf.tbfh.add_tlv(tlvid, parameters)
+
+    def add_credential(self, credential_type, public_key, private_key, cleartext_id):
+        """
+        Add a credential by type to the TBF footer.
+        """
+        for tbf in self._get_tbfs():
+            integrity_blob = tbf.tbfh.get_binary() + tbf.binary
+            tbf.tbff.add_credential(
+                credential_type, public_key, private_key, integrity_blob, cleartext_id
+            )
+
+    def delete_credential(self, credential_type):
+        """
+        Remove a credential by ID from the TBF footer.
+        """
+        for tbf in self._get_tbfs():
+            tbf.tbff.delete_credential(credential_type)
+
+    def verify_credentials(self, public_keys):
+        """
+        Using an optional array of public_key binaries, try to check any
+        contained credentials to verify they are valid.
+        """
+        for tbf in self._get_tbfs():
+            integrity_blob = tbf.tbfh.get_binary() + tbf.binary
+            tbf.tbff.verify_credentials(public_keys, integrity_blob)
+
+    def corrupt_tbf(self, field_name, value):
+        """
+        Modify the TBF root header just before installing the application.
+        """
+        for tbf in self._get_tbfs():
+            tbf.tbfh.corrupt_tbf(field_name, value)
+
     def has_app_binary(self):
         """
         Return true if we have an application binary with this app.
@@ -310,18 +384,15 @@ class TabApp:
         at. This means where the TBF header will go.
         """
 
-        if len(self.tbfs) == 1:
-            tbfh = self.tbfs[0].tbfh
-            app_binary = self.tbfs[0].binary
+        if len(self._get_tbfs()) == 1:
+            tbfh = self._get_tbfs()[0].tbfh
+            app_binary = self._get_tbfs()[0].binary
+            tbff = self._get_tbfs()[0].tbff
 
-            # If the TBF is not compiled for a fixed address, then we can just
-            # use it.
-            if tbfh.has_fixed_addresses() == False:
-                binary = tbfh.get_binary() + app_binary
-
-            else:
+            # If the TBF is compiled for a fixed address, then we make sure the
+            # addresses are lined up.
+            if tbfh.has_fixed_addresses() == True:
                 tbfh.adjust_starting_address(address)
-                binary = tbfh.get_binary() + app_binary
 
             # Check that the binary is not longer than it is supposed to be.
             # This might happen if the size was changed, but any code using this
@@ -329,34 +400,134 @@ class TabApp:
             # the actual binary blob (which should just be padding) to the
             # correct length. If it is too short it is ok, since the board
             # shouldn't care what is in the flash memory the app is not using.
-            binary = self._truncate_binary(binary)
+            binary = self._concatenate_and_truncate_binary(tbfh, app_binary, tbff)
 
             return binary
 
         else:
             raise ("Only valid for one TBF file.")
 
+    def convert(self, format):
+        """
+        Convert a TAB-based app to a different format. Valid formats:
+        - `cbinary`: Create a C struct with a binary representation of the app.
+
+        This is only valid if there is one TBF file.
+        """
+
+        def trim_trailing_zeroes(b):
+            for i in range(len(b) - 1, 0, -1):
+                if b[i] != 0x0:
+                    return b[0 : i + 1]
+
+        def format_data_strings(b):
+            def chunked(source, size):
+                for i in range(0, len(source), size):
+                    yield source[i : i + size]
+
+            out = "{"
+            rows = []
+            for chunk in chunked(b, 10):
+                row = ["0x{:02x}".format(c) for c in chunk]
+                rows.append(", ".join(row))
+            return "{}".format(",\n   ".join(rows))
+
+        if format == "cbinary":
+            output = """
+const unsigned char {name}_embed[] = {{
+   {data}
+}};
+const size_t {name}_embed_size = {length};
+const size_t {name}_embed_actual_size = {trimmed_length};
+"""
+            name = self.get_name()
+            binary = self.get_binary(0)
+            length = len(binary)
+            trimmed = trim_trailing_zeroes(binary)
+            trimmed_length = len(trimmed)
+            data = format_data_strings(trimmed)
+            out = output.format(
+                name=name, length=length, data=data, trimmed_length=trimmed_length
+            )
+            return out
+
     def get_names_and_binaries(self):
         """
         Return (filename, binary) tuples for each contained TBF. This is for
         updating a .tab file.
         """
         out = []
-        for tbf in self.tbfs:
-            binary = tbf.tbfh.get_binary() + tbf.binary
+        for tbf in self._get_tbfs():
             # Truncate in case the header grew and elf2tab padded the binary.
-            binary = self._truncate_binary(binary)
+            binary = self._concatenate_and_truncate_binary(
+                tbf.tbfh, tbf.binary, tbf.tbff
+            )
             out.append((tbf.filename, binary))
         return out
 
+    def _get_tbfs(self):
+        """
+        Helper function so we can implement TBF filtering.
+
+        For normal TBFs (aka PIC TBFs), this doesn't do anything. For fixed
+        address TBFs, this filters the list of TBFs within the TAB to only those
+        that are plausibly within the app memory region for the board.
+        """
+        if self.ram_address_filter == None:
+            return self.tbfs
+        else:
+            tbfs = []
+            for tbf in self.tbfs:
+                fixed_addresses = tbf.tbfh.get_fixed_addresses()
+                if fixed_addresses != None:
+                    if fixed_addresses[0] < self.ram_address_filter or fixed_addresses[
+                        0
+                    ] > (self.ram_address_filter + 0x200000):
+                        pass
+                    else:
+                        tbfs.append(tbf)
+                else:
+                    tbfs.append(tbf)
+            return tbfs
+
+    def _concatenate_and_truncate_binary(self, header, program_binary, footer):
+        size = self.get_size()
+
+        header_binary = header.get_binary()
+        footer_binary = footer.get_binary()
+
+        binary = header_binary + program_binary + footer_binary
+
+        if len(binary) > size:
+            logging.info(
+                "Binary is larger header indicates. Actual:{}, expected:{}".format(
+                    len(binary), size
+                )
+            )
+
+            # First try to shrink the footer to recover the space.
+            extra = len(binary) - size
+            footer.shrink(extra)
+            footer_binary = footer.get_binary()
+            binary = header_binary + program_binary + footer_binary
+
+            # If that was not enough, then try to truncate the actual program
+            # binary, which was likely padded with zeros by elf2tab.
+            if len(binary) > size:
+                raise TockLoaderException(
+                    "Unable to make binary fit size. Compile with larger footer."
+                )
+
+        return binary
+
     def get_crt0_header_str(self):
         """
         Return a string representation of the crt0 header some apps use for
         doing PIC fixups. We assume this header is positioned immediately
         after the TBF header (AKA at the beginning of the application binary).
         """
-        tbfh = self.tbfs[0].tbfh
-        app_binary = self.tbfs[0].binary
+        tbfh = self._get_tbfs()[0].tbfh
+        app_binary = self._get_tbfs()[0].binary
 
         crt0 = struct.unpack("<IIIIIIIIII", app_binary[0:40])
 
@@ -389,35 +560,13 @@ class TabApp:
         """
         out = ""
         out += "Name:                  {}\n".format(self.get_name())
+        out += "Version:               {}\n".format(self.get_app_version())
         out += "Total Size in Flash:   {} bytes\n".format(self.get_size())
 
         if verbose:
-            for tbf in self.tbfs:
+            for tbf in self._get_tbfs():
                 out += textwrap.indent(str(tbf.tbfh), "  ")
         return out
 
-    def _truncate_binary(self, binary):
-        """
-        Optionally truncate binary if the header+protected size has grown, and
-        the actual machine code binary is now too long.
-        """
-        size = self.get_size()
-        if len(binary) > size:
-            logging.info(
-                "Binary is larger than what it says in the header. Actual:{}, expected:{}".format(
-                    len(binary), size
-                )
-            )
-            logging.info("Truncating binary to match.")
-
-            # Check on what we would be removing. If it is all zeros, we
-            # determine that it is OK to truncate.
-            to_remove = binary[size:]
-            if len(to_remove) != to_remove.count(0):
-                raise TockLoaderException("Error truncating binary. Not zero.")
-
-            binary = binary[0:size]
-        return binary
-
     def __str__(self):
         return self.get_name()
