@@ -55,11 +55,17 @@ class BoardInterface:
             "page_size": 4096,
             "no_attribute_table": True,
             "jlink": {
-                "device": "nrf52",
+                "device": "nrf52840_xxAA",
+                "address_maximum": 0x11FFFFFF,
             },
             "openocd": {
                 "cfg": "nordic_nrf52_dk.cfg",
             },
+            "nrfjprog": {
+                "device_family": "NRF52",
+                "qspi_size": 1024 * 1024 * 64,
+                "qspi_address": 0x12000000,
+            },
         },
         "nano33ble": {
             "description": "Arduino Nano 33 BLE board",
@@ -81,7 +87,7 @@ class BoardInterface:
                 "options": ["noreset", "resume"],
                 "commands": {
                     "program": "flash write_image erase {{binary}} {address:#x};\
-			                    verify_image {{binary}} {address:#x};"
+                                verify_image {{binary}} {address:#x};"
                 },
             },
         },
@@ -104,22 +110,24 @@ class BoardInterface:
             "no_attribute_table": True,
             "openocd": {
                 "options": ["nocmdprefix"],
-                "prefix": 'source [find interface/ftdi/digilent-hs1.cfg];\
-		                          ftdi_device_desc \\"Digilent USB Device\\";\
-		                          adapter_khz 10000;\
-		                          transport select jtag;\
-		                          source [find cpld/xilinx-xc7.cfg];\
-		                          source [find cpld/jtagspi.cfg];\
-		                          proc jtagspi_read {{fname offset len}} {{\
-		                            global _FLASHNAME;\
-		                            flash read_bank $_FLASHNAME $fname $offset $len;\
-		                          }};\
-		                          init;\
-		                          jtagspi_init 0 {bitfile};'.format(
+                "prefix": "interface ftdi;\
+                           ftdi_vid_pid 0x0403 0x6010;\
+                           ftdi_channel 0;\
+                           ftdi_layout_init 0x0088 0x008b;\
+                           reset_config none;\
+                           adapter_khz 10000;\
+                           transport select jtag;\
+                           source [find cpld/xilinx-xc7.cfg];\
+                           source [find cpld/jtagspi.cfg];\
+                           proc jtagspi_read {{fname offset len}} {{\
+                             global _FLASHNAME;\
+                             flash read_bank $_FLASHNAME $fname $offset $len;\
+                           }};\
+                           init;\
+                           jtagspi_init 0 {bitfile};".format(
                     bitfile=os.path.join(  # Need path to bscan_spi_xc7a100t.bit
                         os.path.dirname(os.path.realpath(__file__)),
-                        "..",
-                        "bitfiles",
+                        "static",
                         "bscan_spi_xc7a100t.bit",
                     )
                 ),
@@ -132,13 +140,9 @@ class BoardInterface:
         "litex_arty": {
             "description": "LiteX SoC running on an Arty-A7 board",
             "arch": "rv32imc",
-            # Tockloader is currently only supported through the flash file
-            # board interface. The file being operated on is loaded into RAM by
-            # the LiteX bootloader into the main SDRAM. This does the address
-            # translation to the memory-mapped SDRAM bus address.
-            "address_translator": lambda addr: addr - 0x40000000,
             "no_attribute_table": True,
             "flash_file": {
+                "flash_address": 0x40000000,
                 # Set to the maximum RAM size, as the LiteX bootloader will
                 # update the flash image into RAM.
                 "max_size": 0x10000000,
@@ -154,13 +158,20 @@ class BoardInterface:
                 "max_size": 0x00100000,
             },
         },
+        "veer_el2_sim": {
+            "description": "VeeR EL2 running on Verilated simulation",
+            "arch": "rv32imc",
+            "no_attribute_table": True,
+            "flash_file": {
+                "flash_address": 0x20000000,
+            },
+        },
         "qemu_rv32_virt": {
             "description": "QEMU RISC-V 32 bit virt Platform",
             "arch": "rv32imac",
-            # The QEMU-provided binary will be loaded at address 0x80000000
-            "address_translator": lambda addr: addr - 0x80000000,
             "no_attribute_table": True,
             "flash_file": {
+                "flash_address": 0x80000000,
                 # Size of the ROM and PROG region combined, where the resulting
                 # binary will be loaded into by QEMU:
                 "max_size": 0x00200000,
@@ -172,12 +183,7 @@ class BoardInterface:
             "page_size": 2048,
             "no_attribute_table": True,
             "openocd": {
-                "prefix": 'interface hla; \
-                           hla_layout stlink; \
-                           hla_device_desc "ST-LINK/V2-1"; \
-                           hla_vid_pid 0x0483 0x374b; \
-                           set WORKAREASIZE 0xC000; \
-                           source [find target/stm32f3x.cfg];',
+                "cfg": "stm32f3discovery.cfg",
             },
         },
         "stm32f4discovery": {
@@ -186,12 +192,7 @@ class BoardInterface:
             "page_size": 2048,
             "no_attribute_table": True,
             "openocd": {
-                "prefix": 'interface hla; \
-                           hla_layout stlink; \
-                           hla_device_desc "ST-LINK/V2-1"; \
-                           hla_vid_pid 0x0483 0x374b; \
-                           set WORKAREASIZE 0x40000; \
-                           source [find target/stm32f4x.cfg];',
+                "cfg": "stm32f4discovery.cfg",
             },
         },
         "nucleof4": {
@@ -273,6 +274,48 @@ class BoardInterface:
                            source [find target/nrf52.cfg];",
             },
         },
+        "particle_boron": {
+            "description": "nRF52-based cellular enabled development kit",
+            "arch": "cortex-m4",
+            "page_size": 4096,
+            "no_attribute_table": True,
+            "jlink": {
+                "device": "nrf52",
+            },
+        },
+        "opentitan_earlgrey": {
+            "description": "OpenTitan EarlGrey chip targeting the Earlgrey-M2.5.2-RC0 hardware",
+            "arch": "rv32imc",
+            "page_size": 512,
+            "no_attribute_table": True,
+            "flash_file": {
+                "flash_address": 0x20000000,
+                # Set to the maximum flash size.
+                "max_size": 0x00100000,
+            },
+        },
+        "opentitan_earlgrey_secure_boot": {
+            "description": "OpenTitan EarlGrey chip targeting the Earlgrey-M2.5.2-RC0 hardware with secure boot",
+            "arch": "rv32imc",
+            "page_size": 512,
+            "no_attribute_table": True,
+            "flash_file": {
+                "flash_address": 0x20010000,
+                # Set to the half of maximum flash size, to keep image sizes smaller
+                # (also ensures room for data at end of flash).
+                "max_size": 0x00070000,
+            },
+        },
+        "cy8cproto_62_4343_w": {
+            "description": "Infineon board based on the PSoC 62xA SoC",
+            "arch": "cortex-m0",
+            "no_attribute_table": True,
+            "probers": {"chip": "CY8C624AAZI-S2D44"},
+            "flash_file": {
+                "flash_address": 0x10000000,
+                "flush_command": "probe-rs download {binary} --binary-format bin --base-address 0x10000000 --chip CY8C624AAZI-S2D44",
+            },
+        },
     }
 
     def __init__(self, args):
@@ -334,6 +377,12 @@ class BoardInterface:
         # methods need. There may be flags specific to a particular
         # communication interface.
 
+    def is_known_board(board):
+        """
+        Check if the board name is a known board.
+        """
+        return board in BoardInterface.KNOWN_BOARDS.keys()
+
     def translate_address(self, address):
         """
         Translate an address from MCU address space to the address required for
@@ -470,6 +519,27 @@ class BoardInterface:
         # re-implemented.
         raise TockLoaderException("No bootloader, cannot set start address.")
 
+    def get_apps_start_address(self):
+        """
+        Return the address in flash where apps start.
+        """
+        if hasattr(self, "app_address") and self.app_address:
+            return self.app_address
+        else:
+            attributes = self.get_all_attributes()
+            for attribute in attributes:
+                if attribute and attribute["key"] == "appaddr":
+                    return int(attribute["value"], 0)
+
+        # Or, if we don't know, return None
+        return None
+
+    def get_flash_address(self):
+        """
+        Return the address where flash starts.
+        """
+        return None
+
     def _decode_attribute(self, raw):
         try:
             key = raw[0:8].decode("utf-8").strip(bytes([0]).decode("utf-8"))
