@@ -17,11 +17,16 @@ import sys
 import time
 import threading
 
-# Although Windows is not supported actively, this allow features that "just
-# work" to work on Windows.
+# Windows support in tockloader is currently experimental. Please report bugs,
+# and ideally rough paths to fixing them. The core maintainers have limited
+# access to Windows machines, and support in testing fixes/updates is helpful.
 if platform.system() != "Windows":
     import fcntl
 
+    _IS_WINDOWS = False
+else:
+    _IS_WINDOWS = True
+
 import serial
 import serial.tools.list_ports
 import serial.tools.miniterm
@@ -165,7 +170,7 @@ class BootloaderSerial(BoardInterface):
                 index = helpers.menu(
                     ports,
                     return_type="index",
-                    title="Multiple serial port options found. Which would you like to use?",
+                    title="Multiple matching serial port options found. Which would you like to use?",
                 )
         else:
             # Just find any port. If one, use that. If multiple, ask user.
@@ -173,27 +178,104 @@ class BootloaderSerial(BoardInterface):
             # Macs will report Bluetooth devices with serial, which is
             # almost certainly never what you want, so drop those.
             ports = [p for p in ports if "Bluetooth-Incoming-Port" not in p.device]
+            ports = [p for p in ports if "cu.BLTH" not in p.device]
+            index = None
 
             if len(ports) == 0:
                 raise TockLoaderException(
                     "No serial ports found. Is the board connected?"
                 )
 
-            logging.info(
-                'No serial port with device name "{}" found.'.format(device_name)
-            )
-            logging.info(
-                "Found {} serial port{}.".format(len(ports), ("s", "")[len(ports) == 1])
-            )
+            # Attempt to workaround the issue with the nRF52840dk (PCA10056),
+            # particularly newer revisions of that board, that open two serial
+            # ports by detecting which one is the correct port.
+            #
+            # We first try to only run this if we think that the user actually
+            # has a nRF52840dk plugged in. We check that by looking for two
+            # devices that both have "J-Link - CDC" in the name.
+            #
+            # If we find that, we use the `nrfjprog --com` command which lists
+            # attached ports and their VCOM indices. We want VCOM0. We use the
+            # pynrfjprog to run the same operation as `--com`. `nrfjprog --com`
+            # has output that looks like:
+            #
+            # ```
+            # $ nrfjprog --com
+            # 1050288520    /dev/tty.usbmodem0010502885201    VCOM0
+            # 1050288520    /dev/tty.usbmodem0010502885203    VCOM1
+            # ```
+            jlink_cdc_ports = [p for p in ports if "J-Link - CDC" in p.description]
+            if len(jlink_cdc_ports) == 2:
+                # It looks like the user has the nRF52840dk connected.
+                try:
+                    import pynrfjprog
+                    from pynrfjprog import LowLevel
 
-            if len(ports) == 1 or any:
-                index = 0
-            else:
-                index = helpers.menu(
-                    ports,
-                    return_type="index",
-                    title="Multiple serial port options found. Which would you like to use?",
+                    api = pynrfjprog.LowLevel.API()
+                    if not api.is_open():
+                        api.open()
+
+                    vcom0_path = None
+                    jtag_emulators = api.enum_emu_con_info()
+                    for jtag_emulator in jtag_emulators:
+                        jtag_emulator_ports = api.enum_emu_com_ports(
+                            jtag_emulator.serial_number
+                        )
+                        for jtag_emulator_port in jtag_emulator_ports:
+                            # We want to see VCOM == 0
+                            if jtag_emulator_port.vcom == 0:
+                                vcom0_path = jtag_emulator_port.path
+                                break
+                        # Only support one connected nRF52840dk for now.
+                        break
+
+                    if vcom0_path != None:
+                        # On mac, the nrfjprog tool uses the /dev/tty* paths,
+                        # and we need the /dev/cu* paths. We just hack in a
+                        # substitution here which will only have an effect on
+                        # the mac paths.
+                        vcom0_path_standarized = vcom0_path.replace(
+                            "/dev/tty.usbmodem", "/dev/cu.usbmodem"
+                        )
+
+                        # Update list of ports to just the one we found for
+                        # VCOM0.
+                        ports = [p for p in ports if vcom0_path_standarized in p.device]
+                        index = 0
+                        logging.info(
+                            'Discovered "{}" as nRF52840dk VCOM0.'.format(
+                                vcom0_path_standarized
+                            )
+                        )
+
+                    # Must close this to end the underlying pynrfjprog process.
+                    # Otherwise on my machine it sits at 100% CPU.
+                    api.close()
+                except:
+                    # Any error with nrfjprog we just don't use this
+                    # optimization.
+                    pass
+
+            # Continue searching if our special-case discovery did not find
+            # anything.
+            if index == None:
+                logging.info(
+                    'No serial port with device name "{}" found.'.format(device_name)
                 )
+                logging.info(
+                    "Found {} serial port{}.".format(
+                        len(ports), ("s", "")[len(ports) == 1]
+                    )
+                )
+
+                if len(ports) == 1 or any:
+                    index = 0
+                else:
+                    index = helpers.menu(
+                        ports,
+                        return_type="index",
+                        title="Multiple serial port options found. Which would you like to use?",
+                    )
 
         # Choose port. This should be a serial.ListPortInfo type.
         port = ports[index]
@@ -201,9 +283,11 @@ class BootloaderSerial(BoardInterface):
         logging.info('Using "{}".'.format(port))
 
         # Save the serial number. This might help us reconnect later if say we
-        # have to boot into the bootloader and the OS assigns a new port name
-        # to the same physical board.
-        self.sp_serial_number = port.serial_number
+        # have to boot into the bootloader and the OS assigns a new port name to
+        # the same physical board. On _some_ USB devices there is no serial
+        # number. This was first found on the WM1110-dev board. In that case we
+        # fall back the port name.
+        self.sp_serial_number = port.serial_number or port.name
 
         # Improve UI for users
         helpers.set_terminal_title_from_port_info(port)
@@ -296,21 +380,36 @@ class BootloaderSerial(BoardInterface):
         # Before connecting, check whether there is another tockloader process
         # running, and if it's a listen, pause that listen (unless we are also
         # doing a listen), otherwise bail out.
-        self.comm_path = "/tmp/tockloader." + self._get_serial_port_hash()
-        if os.path.exists(self.comm_path):
-            # Open a socket to the other tockloader instance if one exists.
-            self.client_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+
+        # Windows has only partial unix socket support, so Python rejects them.
+        # Work around this by listening on a reasonably-unlikely-to-collide
+        # localhost port derived from the serial port name.
+        if _IS_WINDOWS:
+            self.comm_port = self._get_serial_port_hashed_to_ip_port()
+            self.client_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
             try:
-                self.client_sock.connect(self.comm_path)
+                self.client_sock.connect(("localhost", self.comm_port))
+                logging.debug("Connected to existing `tockloader listen`")
             except ConnectionRefusedError:
-                logging.warning("Found stale tockloader server, removing.")
-                logging.warning(
-                    "This may occur if a previous tockloader instance crashed."
+                logging.debug(
+                    f"No other listen instances running (tried localhost::{self.comm_port})"
                 )
-                os.unlink(self.comm_path)
                 self.client_sock = None
         else:
-            self.client_sock = None
+            self.comm_path = "/tmp/tockloader." + self._get_serial_port_hash()
+            if os.path.exists(self.comm_path):
+                self.client_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+                try:
+                    self.client_sock.connect(self.comm_path)
+                except ConnectionRefusedError:
+                    logging.warning("Found stale tockloader server, removing.")
+                    logging.warning(
+                        "This may occur if a previous tockloader instance crashed."
+                    )
+                    os.unlink(self.comm_path)
+                    self.client_sock = None
+            else:
+                self.client_sock = None
 
         # Check if another tockloader instance exists based on whether we were
         # able to create a socket to it.
@@ -333,7 +432,6 @@ class BootloaderSerial(BoardInterface):
 
             r = ""
             while True:
-
                 while "\n" not in r:
                     r += self.client_sock.recv(100).decode("utf-8")
 
@@ -386,17 +484,23 @@ class BootloaderSerial(BoardInterface):
             # listen, allow the other tockloader instance to complete, and then
             # resume listening.
 
-            # Create the socket we will listen on.
-            self.server_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+            if _IS_WINDOWS:
+                self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                self.server_sock.bind(("localhost", self.comm_port))
+                logging.debug(f"listening on localhost::{self.comm_port}")
+            else:
+                # Create the socket we will listen on.
+                self.server_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
 
-            # Close the file descriptor if exec() is called (apparently). I'm
-            # not sure why we need this (or if we do).
-            flags = fcntl.fcntl(self.server_sock, fcntl.F_GETFD)
-            fcntl.fcntl(self.server_sock, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
+                # Close the file descriptor if exec() is called (apparently). I'm
+                # not sure why we need this (or if we do).
+                flags = fcntl.fcntl(self.server_sock, fcntl.F_GETFD)
+                fcntl.fcntl(self.server_sock, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
+
+                self.server_sock.bind(self.comm_path)
 
             # Finish setting up the socket, and spawn a thread to listen on that
             # socket.
-            self.server_sock.bind(self.comm_path)
             self.server_sock.listen(1)
             self.server_event = threading.Event()
             self.server_thread = threading.Thread(
@@ -411,7 +515,8 @@ class BootloaderSerial(BoardInterface):
             def server_cleanup():
                 if self.server_sock is not None:
                     self.server_sock.close()
-                    os.unlink(self.comm_path)
+                    if not _IS_WINDOWS:
+                        os.unlink(self.comm_path)
 
             atexit.register(server_cleanup)
 
@@ -515,6 +620,14 @@ class BootloaderSerial(BoardInterface):
         """
         return hashlib.sha1(self.sp.port.encode("utf-8")).hexdigest()
 
+    def _get_serial_port_hashed_to_ip_port(self):
+        """
+        This is a bit of a hack, but it's means to find a reasonably unlikely
+        to collide port number based on the serial port used to talk to the
+        board.
+        """
+        return int(self._get_serial_port_hash(), 16) % 40000 + 10000
+
     def _toggle_bootloader_entry_DTR_RTS(self):
         """
         Use the DTR and RTS lines on UART to reset the chip and assert the
@@ -896,9 +1009,35 @@ class BootloaderSerial(BoardInterface):
                     "Padding binary with {} bytes already on chip.".format(remaining)
                 )
 
+        # Get indices of pages that have valid data to write.
+        valid_pages = []
+        for i in range(len(binary) // self.page_size):
+            if not all(
+                b == 0 for b in binary[i * self.page_size : (i + 1) * self.page_size]
+            ):
+                valid_pages.append(i)
+
+        # Make sure that there is at least one valid page. If we are only trying
+        # to write 0s then all pages would have been removed. In that case, we
+        # want to write all requested pages.
+        if len(valid_pages) == 0:
+            for i in range(len(binary) // self.page_size):
+                valid_pages.append(i)
+
+        # Make sure to always include one blank page (if exists) after the end
+        # of a valid page. There might be a usable 0 on the next page. It's
+        # unlikely there is more than entire page of valid 0s on the next page.
+        ending_pages = []
+        for i in valid_pages:
+            if (not (i + 1) in valid_pages) and (
+                (i + 1) < (len(binary) // self.page_size)
+            ):
+                ending_pages.append(i + 1)
+        valid_pages = valid_pages + ending_pages
+
         # Loop through the binary by pages at a time until it has been flashed
         # to the chip.
-        for i in tqdm(range(len(binary) // self.page_size)):
+        for i in tqdm(valid_pages):
             # Create the packet that we send to the bootloader. First four
             # bytes are the address of the page.
             pkt = struct.pack("<I", address + (i * self.page_size))
@@ -938,7 +1077,7 @@ class BootloaderSerial(BoardInterface):
                 )
 
         # And check the CRC
-        self._check_crc(address, binary)
+        self._check_crc(address, binary, valid_pages)
 
     def read_range(self, address, length):
         # Can only read up to 4095 bytes at a time.
@@ -1054,29 +1193,66 @@ class BootloaderSerial(BoardInterface):
 
         return crc
 
-    def _check_crc(self, address, binary):
+    def _check_crc(self, address, binary, valid_pages):
         """
         Compares the CRC of the local binary to the one calculated by the
         bootloader.
         """
-        # Check the CRC
-        crc_data = self._get_crc_internal_flash(address, len(binary))
 
-        # Now interpret the returned bytes as the CRC
-        crc_bootloader = struct.unpack("<I", crc_data[0:4])[0]
+        def get_sequences(l):
+            """
+            Find the start and end of sequences in `l`. [0, 1, 2, 5, 6, 11]
+            would return: [(0, 3), (5, 7), (11, 12)].
+            """
+            sequences = []
+            start = -1
+            last = -1
+            for i in l:
+                if start == -1:
+                    start = i
+                else:
+                    if i - 1 == start or i - 1 == last:
+                        last = i
+                    else:
+                        if last == -1:
+                            last = start + 1
+                        sequences.append((start, last))
+                        start = i
+                        last = -1
+            sequences.append((start, l[-1] + 1))
+            return sequences
 
-        # Calculate the CRC locally
-        crc_function = crcmod.mkCrcFun(0x104C11DB7, initCrc=0, xorOut=0xFFFFFFFF)
-        crc_loader = crc_function(binary, 0)
+        # Compute the CRC for each stretch of pages in the flashed binary.
+        crcs = []
+        for start, last in get_sequences(valid_pages):
+            # Found end of a run. Check CRC.
+            crc_address = address + (start * self.page_size)
+            crc_length = (last - start) * self.page_size
 
-        if crc_bootloader != crc_loader:
-            raise TockLoaderException(
-                "Error: CRC check failed. Expected: 0x{:04x}, Got: 0x{:04x}".format(
-                    crc_loader, crc_bootloader
-                )
+            # Check the CRC
+            crc_data = self._get_crc_internal_flash(crc_address, crc_length)
+
+            # Now interpret the returned bytes as the CRC
+            crc_bootloader = struct.unpack("<I", crc_data[0:4])[0]
+
+            # Calculate the CRC locally
+            crc_function = crcmod.mkCrcFun(0x104C11DB7, initCrc=0, xorOut=0xFFFFFFFF)
+            crc_loader = crc_function(
+                binary[start * self.page_size : last * self.page_size], 0
             )
-        else:
-            logging.info("CRC check passed. Binaries successfully loaded.")
+
+            # Add to list of crcs to compare
+            crcs.append((crc_bootloader, crc_loader))
+
+        # Check each run of pages.
+        for crc_bootloader, crc_loader in crcs:
+            if crc_bootloader != crc_loader:
+                raise TockLoaderException(
+                    "Error: CRC check failed. Expected: 0x{:04x}, Got: 0x{:04x}".format(
+                        crc_loader, crc_bootloader
+                    )
+                )
+        logging.info("CRC check passed. Binaries successfully loaded.")
 
     def get_attribute(self, index):
         # Check for cached value.
