@@ -19,7 +19,7 @@ import tempfile
 import time
 
 from .board_interface import BoardInterface
-from .exceptions import TockLoaderException
+from .exceptions import TockLoaderException, ChannelAddressErrorException
 
 
 class JLinkExe(BoardInterface):
@@ -36,6 +36,15 @@ class JLinkExe(BoardInterface):
             if platform.system() == "Windows":
                 self.jlink_cmd = "JLink"
 
+        # Obtain serial number if --jlink-serial-number argument provided
+        self.jlink_serial_number = getattr(self.args, "jlink_serial_number")
+
+        # By default we assume that jlinkexe can be used to read any address on
+        # this board, so we set `address_maximum` to None. In some cases,
+        # however, particularly with external flash chips, jlinkexe may not be
+        # able to read all flash addresses tockloader needs to access.
+        self.address_maximum = None
+
     def attached_board_exists(self):
         # Get a list of attached jlink devices, check if that list has at least
         # one entry.
@@ -59,10 +68,17 @@ class JLinkExe(BoardInterface):
             # there is only one for now.
             emulator = emulators[0]
             # Check for known JTAG board.
+            if emulator["ProductName"] == "J-Link (unknown)":
+                logging.warning("JLink EmuList reports 'unknown' product.")
+                logging.warning("Please try upgrading JLinkExe to latest version.")
             if emulator["ProductName"] == "J-Link OB-SAM3U128-V2-NordicSem":
                 # This seems to match both the nRF52dk (PCA10040) and the
                 # nRF52840dk (PCA10056). From a jlink perspective, they are
-                # the same, which is nice.
+                # close enough, which is nice.
+                return "nrf52dk"
+            if emulator["ProductName"] == "J-Link OB-nRF5340-NordicSemi":
+                # This matches the new (3.0.0+) version of the nRF52840dk
+                # (PCA10056).
                 return "nrf52dk"
             if emulator["ProductName"] == "J-Link" and emulator[
                 "Serial number"
@@ -116,6 +132,8 @@ class JLinkExe(BoardInterface):
                 self.jlink_if = board["jlink"]["if"]
             if self.jlink_speed == None and "speed" in board["jlink"]:
                 self.jlink_speed = board["jlink"]["speed"]
+            if self.address_maximum == None and "address_maximum" in board["jlink"]:
+                self.address_maximum = board["jlink"]["address_maximum"]
 
             # And we may need to setup other common board settings.
             self._configure_from_known_boards()
@@ -181,6 +199,10 @@ class JLinkExe(BoardInterface):
                 jlink_file.name,
             )
 
+            # Append target selector if serial number provided.
+            if self.jlink_serial_number:
+                jlink_command += " -USB {}".format(self.jlink_serial_number)
+
             logging.debug('Running "{}".'.format(jlink_command))
 
             def print_output(subp):
@@ -230,7 +252,8 @@ class JLinkExe(BoardInterface):
             # Cleanup files on Windows if needed.
             if not self.args.debug:
                 os.remove(jlink_file.name)
-                os.remove(temp_bin.name)
+                if binary or not write:
+                    os.remove(temp_bin.name)
 
             return ret
 
@@ -311,14 +334,16 @@ class JLinkExe(BoardInterface):
             # Cleanup files on Windows if needed.
             if not self.args.debug:
                 os.remove(jlink_file.name)
-                os.remove(temp_bin.name)
 
         return emulators
 
-    def flash_binary(self, address, binary):
+    def flash_binary(self, address, binary, pad=False):
         """
         Write using JTAG
         """
+        if self.address_maximum and address > self.address_maximum:
+            raise ChannelAddressErrorException()
+
         # Make sure we respect page boundaries in case the chip and jlink
         # implementation will only work correctly when writing entire pages.
         address, binary = self._align_and_stretch_to_page(address, binary)
@@ -333,6 +358,9 @@ class JLinkExe(BoardInterface):
         self._run_jtag_commands(commands, binary)
 
     def read_range(self, address, length):
+        if self.address_maximum and address > self.address_maximum:
+            raise ChannelAddressErrorException()
+
         commands = []
         if self.jlink_device == "cortex-m0":
             # We are in generic mode, trying to read attributes.
@@ -370,6 +398,9 @@ class JLinkExe(BoardInterface):
         return read
 
     def clear_bytes(self, address):
+        if self.address_maximum and address > self.address_maximum:
+            raise ChannelAddressErrorException()
+
         logging.debug("Clearing 512 bytes starting at address {:#0x}".format(address))
 
         # Write 512 bytes of 0xFF as that seems to work.
@@ -432,9 +463,19 @@ class JLinkExe(BoardInterface):
             return
 
         logging.status("Starting JLinkExe JTAG connection.")
+
+        # Include serial number specifier if `--jlink-serial-number` flag provided.
+        jlink_serial_number_str = ""
+        if self.jlink_serial_number:
+            jlink_serial_number_str = ("-USB {}").format(self.jlink_serial_number)
+
         jtag_p = subprocess.Popen(
-            "{} -device {} -if {} -speed {} -autoconnect 1 -jtagconf -1,-1".format(
-                self.jlink_cmd, self.jlink_device, self.jlink_if, self.jlink_speed
+            "{} -device {} -if {} -speed {} -autoconnect 1 {} --jtagconf -1,-1".format(
+                self.jlink_cmd,
+                self.jlink_device,
+                self.jlink_if,
+                self.jlink_speed,
+                jlink_serial_number_str,
             ).split(),
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
