@@ -31,6 +31,9 @@ class OpenOCD(BoardInterface):
         # Command can be passed in as an argument, otherwise use default.
         self.openocd_cmd = getattr(self.args, "openocd_cmd")
 
+        # Store the serial number if provided
+        self.openocd_serial_number = getattr(self.args, "openocd_serial_number")
+
     def attached_board_exists(self):
         # Get a list of attached devices, check if that list has at least
         # one entry.
@@ -132,8 +135,14 @@ class OpenOCD(BoardInterface):
         # Defaults.
         prefix = ""
         source = "source [find board/{board}];".format(board=self.openocd_board)
+
         cmd_prefix = "init; reset init; halt;"
         cmd_suffix = ""
+        serial_no_cmd = ""
+
+        # Add serial number specification if provided
+        if hasattr(self, "openocd_serial_number") and self.openocd_serial_number:
+            serial_no_cmd = "adapter serial {};".format(self.openocd_serial_number)
 
         # Do the customizations
         if "workareazero" in self.openocd_options:
@@ -151,12 +160,15 @@ class OpenOCD(BoardInterface):
         if exit:
             cmd_suffix += "exit"
 
-        command_param = "{prefix} {source} {cmd_prefix} {cmd} {cmd_suffix}".format(
-            prefix=prefix,
-            source=source,
-            cmd_prefix=cmd_prefix,
-            cmd="; ".join(commands),
-            cmd_suffix=cmd_suffix,
+        command_param = (
+            "{prefix} {serial_no_cmd} {source} {cmd_prefix} {cmd} {cmd_suffix}".format(
+                prefix=prefix,
+                serial_no_cmd=serial_no_cmd,
+                source=source,
+                cmd_prefix=cmd_prefix,
+                cmd="; ".join(commands),
+                cmd_suffix=cmd_suffix,
+            )
         )
 
         return (
@@ -172,7 +184,7 @@ class OpenOCD(BoardInterface):
             [commands], binary, write
         )
 
-        logging.debug('Running "{}".'.format(openocd_command.replace("$", "\$")))
+        logging.debug('Running "{}".'.format(openocd_command.replace("$", "\\$")))
 
         def print_output(subp):
             response = ""
@@ -231,6 +243,11 @@ You may need to update OpenOCD to the version in latest git master."
                 openocd_cmd=self.openocd_cmd
             )
         )
+        openocd_commands.append(
+            '{openocd_cmd} -c "source [find interface/stlink.cfg]; transport select hla_swd; source [find target/stm32f4x.cfg]; init; exit;"'.format(
+                openocd_cmd=self.openocd_cmd
+            )
+        )
 
         # These are the magic strings in the output of openocd we are looking
         # for. If there is a better way to do this then we should change. But,
@@ -238,8 +255,10 @@ You may need to update OpenOCD to the version in latest git master."
         # openocd output, board is the name in the known boards struct.
         magic_strings_boards = [
             ("J-Link OB-SAM3U128-V2-NordicSemi", "nrf52dk"),
+            ("J-Link OB-nRF5340-NordicSemi", "nrf52dk"),
             ("(mfg: 0x049 (Xilinx), part: 0x3631, ver: 0x1)", "arty"),
             ("SWD DPIDR 0x2ba01477", "microbit_v2"),
+            ("stm32f4x.cpu", "stm32f4discovery"),
         ]
 
         emulators = []
@@ -280,12 +299,12 @@ You may need to update OpenOCD to the version in latest git master."
 
         return emulators
 
-    def flash_binary(self, address, binary):
+    def flash_binary(self, address, binary, pad=False):
         """
         Write using openocd `program` command.
         """
         # The "normal" flash command uses `program`.
-        command = "program {{binary}} verify {address:#x};"
+        command = "program {{binary}} verify {address:#x}; reset;"
 
         # Check if the configuration wants to override the default program command.
         if "program" in self.openocd_commands:
@@ -369,15 +388,16 @@ You may need to update OpenOCD to the version in latest git master."
         self._configure_from_known_boards()
 
         # Check that we learned what we needed to learn.
-        if (
-            self.board == None
-            or self.arch == None
-            or self.openocd_board == "cortex-m0"
-            or self.page_size == 0
-        ):
+        if self.board is None:
+            raise TockLoaderException("Could not determine the current board")
+        if self.arch is None:
+            raise TockLoaderException("Could not determine the current arch")
+        if self.openocd_board == "cortex-m0":
             raise TockLoaderException(
-                "Could not determine the current board or arch or openocd board name"
+                "Could not determine the current openocd board name"
             )
+        if self.page_size == 0:
+            raise TockLoaderException("Could not determine the current page size")
 
     def run_terminal(self):
         self.open_link_to_board()
@@ -394,7 +414,7 @@ You may need to update OpenOCD to the version in latest git master."
             exit=False,
         )
 
-        logging.debug('Running "{}".'.format(openocd_command.replace("$", "\$")))
+        logging.debug('Running "{}".'.format(openocd_command.replace("$", "\\$")))
 
         cleanup = []
         try:
